"use strict"

/* This is a Node script, not browser javascript.
   This script makes a collision mask from spritesheet.png so sprites can have
   pixel-perfect collision. You only need to rerun it if you've edited
   the spritesheet.
*/

const fs=require('fs');
/* Since PNGReader's already in this directory in a browserified package,
   requiring it as an outside npm dependency would be redundant. */
const PNGReader=require("./pngreader_browser.js")

new PNGReader(fs.readFileSync("spritesheet.png")).parse(
 (err,png) => {
  if(err) {
   throw(err);
  }
  // Make the collision mask RLE. Start at collision-false state, count how
  // many false pixels there are, then how many true pixels, then how many
  // false, etc. to get a list of integers. 
  var runLength=0;
  var state=false;
  var runs=[];
  for(var y=0;y<png.height;++y) {
   for(var x=0;x<png.width;++x) {
    var alpha=png.getPixel(x,y)[3];
    var newState=(alpha>0);
    if(state!=newState) {
     runs.push(runLength);
     state=newState;
     runLength=1;
    }
    else {
     ++runLength;
    }
   }
  }
  runs.push(runLength);

  var outputObject={
   width:png.width,
   height:png.height,
   rle:runs,
  }
  var outputString=
      '/* This file was output from generate_spritesheet_mask.js */\n'+
      'const SPRITESHEET_MASK='+JSON.stringify(outputObject)+";\n";
  
  fs.writeFileSync("spritesheet_mask.js",outputString);
  console.log("Successfully wrote spritesheet_mask.js.");
 }
);
