"use strict"

/*
  Most game sprites are subclasses of CustomSprite that just call configuration
  methods in their constructors; see MODDING.txt for documentation.
  Please remember: for a sprite's collision detection to work, you have to
  run generate_spritesheet_mask.js in Node after editing the spritesheet.
*/

class CustomSprite {
 constructor() {
  this.layer=1 // in front of NoncustomSprites by default
  this.periodicSpawns=[]
  this.periodicSpawnCycleCounter=0;
  this.blinkFrames=0;
 }
 setRect(left,top,width,height) {
  this.left=left|0
  this.top=top|0
  this.width=width|0
  this.height=height|0
  return this;
 }
 setAnimation(spritesheetLeft, spritesheetTop, spritesheetGrid,
	      animationFrameCount, gameFramesPerAnimationFrame) {
  this.spritesheetLeft=spritesheetLeft|0;
  this.spritesheetTop=spritesheetTop|0;
  this.spritesheetGrid=spritesheetGrid|0;
  this.animationFrameCount=animationFrameCount|0;
  this.gameFramesPerAnimationFrame=gameFramesPerAnimationFrame|0;
  this.animationCycleCounter=0;
  this.texLeft=this.spritesheetLeft;
  this.texTop=this.spritesheetTop;
  return this;
 }
 setLayer(layer) {
  this.layer=layer|0
  return this;
 }
 setXSinusoidMove(range, initialPhase, gameFramesPerLoop) {
  this.xSineRange=range|0;
  this.xSineInitialPhase=initialPhase|0;
  this.xSineGameFramesPerLoop=gameFramesPerLoop|0;
  this.xSineInitialLeft=this.left|0;
  this.xSinePhaseCycle=0;
  if(this.moveType=="ySine") { this.moveType="xySine"; }
  else { this.moveType="xSine"; }
  return this;
 }
 setYSinusoidMove(range, initialPhase, gameFramesPerLoop) {
  this.ySineRange=range|0;
  this.ySineInitialPhase=initialPhase|0;
  this.ySineGameFramesPerLoop=gameFramesPerLoop|0;
  this.ySineInitialTop=this.top|0;
  this.ySinePhaseCycle=0;
  if(this.moveType=="xSine") { this.moveType="xySine"; }
  else { this.moveType="ySine"; }
  return this;
 }
 setPatrolMove(xVelocityNumerator, yVelocityNumerator, speedDenominator,
	       altSpritesheetTop) {
  this.xVelocityNumerator=xVelocityNumerator|0;
  this.yVelocityNumerator=yVelocityNumerator|0;
  this.speedDenominator=speedDenominator|0;
  this.moveType="patrol";
  this.xSubpixels=0; this.ySubpixels=0;
  if(altSpritesheetTop!=undefined) {
   this.altSpritesheetTop=altSpritesheetTop|0;
  }
  return this;
 }
 setXWarpMove(xVelocityNumerator, speedDenominator) {
  this.xVelocityNumerator=xVelocityNumerator|0;
  this.yVelocityNumerator=0;
  this.speedDenominator=speedDenominator|0;
  this.moveType="warp";  
  this.xSubpixels=0; this.ySubpixels=0;
  return this;
 }
 setYWarpMove(yVelocityNumerator, speedDenominator) {
  this.xVelocityNumerator=0;
  this.yVelocityNumerator=yVelocityNumerator|0;  
  this.speedDenominator=speedDenominator|0;
  this.moveType="warp";  
  this.xSubpixels=0; this.ySubpixels=0;
  return this;
 } 
 setProjectileMove(xVelocityNumerator, yVelocityNumerator,
		   xAccelerationNumerator, yAccelerationNumerator,
		   speedDenominator) {
  this.xVelocityNumerator=xVelocityNumerator|0;
  this.yVelocityNumerator=yVelocityNumerator|0;
  this.xAccelerationNumerator=xAccelerationNumerator|0;
  this.yAccelerationNumerator=yAccelerationNumerator|0;
  this.speedDenominator=speedDenominator|0;
  this.moveType="projectile";
  this.xSubpixels=0; this.ySubpixels=0;
  return this;
 }
 setAgeLimit(frames) {
  this.expirationFrames=frames;
  return this;
 }
 setDespawnAnimation(spritesheetLeft, spritesheetTop, spritesheetGrid,
	      animationFrameCount, gameFramesPerAnimationFrame) {
  this.despawnSpritesheetLeft=spritesheetLeft|0;
  this.despawnSpritesheetTop=spritesheetTop|0;
  this.despawnSpritesheetGrid=spritesheetGrid|0;
  this.despawnAnimationFrameCount=animationFrameCount|0;
  this.gameFramesPerDespawnAnimationFrame=gameFramesPerAnimationFrame|0;
  return this;
 }
 addPeriodicSpawn(spriteClass, intervalFrames, offsetFrames) {
  this.periodicSpawns.push(
   {
    spriteClass: spriteClass,
    intervalFrames: intervalFrames|0,
    offsetFrames: offsetFrames|0,
   }
  )
  return this;
 }
 setPeriodicPause(activeFrames,waitFrames,
		  waitSpritesheetLeft, waitSpritesheetTop) {
  this.activeFrames=activeFrames|0
  this.waitFrames=waitFrames|0;
  this.activityPhase="active";
  this.framesInActivityPhase=0;
  if(waitSpritesheetLeft!=undefined) {
   this.waitSpritesheetLeft=waitSpritesheetLeft|0;
  }
  if(waitSpritesheetTop!=undefined) {
   this.waitSpritesheetTop=waitSpritesheetTop|0;
  }   
  return this;
 }
 suppressBlink() {
  this.blinkSuppressed=true;
  return this;
 }
 runFrame() {
  if(this.activityPhase=="wait") {
   ++this.framesInActivityPhase;
   if(this.framesInActivityPhase>=this.waitFrames) {
    this.activityPhase="active";
    this.framesInActivityPhase=0;
    // next frame will be when it actually acts, and advanceAnimation
    // there will pick up its graphic where it left off
   }
  }
  else {
   this.runActiveFrame();
   if(this.waitFrames) {
    ++this.framesInActivityPhase;
    if(this.framesInActivityPhase>=this.activeFrames) {
     this.activityPhase="wait";
     this.framesInActivityPhase=0;
     if("waitSpritesheetLeft" in this) {
      this.texLeft=this.waitSpritesheetLeft;
     }
     if("waitSpritesheetTop" in this) {
      this.texTop=this.waitSpritesheetTop;
     }     
    }
   }
  }  
  if(this.blinkFrames>0) { --this.blinkFrames; }
 }
 runActiveFrame() {
  this.handleExpiration();
  this.advanceAnimation();
  this.runMovement();
  this.handlePeriodicSpawns();
 }
 handleExpiration() {
  if(this.expirationFrames) {
   --this.expirationFrames;
   if(this.expirationFrames<=0) {
    this.despawned=true;
    return;
   }
  }
 }
 advanceAnimation() {
  ++this.animationCycleCounter;
  if(this.gameFramesPerAnimationFrame<1) { var whichFrame=0; }
  else {
   var whichFrame = Math.floor(
    this.animationCycleCounter/this.gameFramesPerAnimationFrame
   );
  }
  if(whichFrame>=this.animationFrameCount && this.moveType=="despawn") {
   this.despawned=true;
  }
  else {
   whichFrame=whichFrame%this.animationFrameCount;
   this.texLeft=this.spritesheetLeft+this.spritesheetGrid*whichFrame;
  }
 }
 runMovement() {
  switch(this.moveType) {
  case "patrol": this.runSubpixelMovement("patrol"); break;
  case "projectile": this.runSubpixelMovement("projectile"); break;
  case "warp": this.runSubpixelMovement("warp"); break;
  case "xSine": this.runXSineMovement(); break;
  case "ySine": this.runYSineMovement(); break;
  case "xySine":
   this.runXSineMovement();
   this.runYSineMovement();
   break;
   // undefined and despawn movement have nothing to run here
  }
 }
 runSubpixelMovement(collisionStyle) {
  if(this.xAccelerationNumerator) {
   this.xVelocityNumerator+=this.xAccelerationNumerator;
  }
  if(this.yAccelerationNumerator) {
   this.yVelocityNumerator+=this.yAccelerationNumerator;
  }
  this.xSubpixels+=this.xVelocityNumerator;
  this.ySubpixels+=this.yVelocityNumerator;
  while(Math.abs(this.xSubpixels)>=this.speedDenominator ||
	Math.abs(this.ySubpixels)>=this.speedDenominator) {
   var xStep=0, yStep=0;
   if(this.xSubpixels>=this.speedDenominator) {
    this.xSubpixels-=this.speedDenominator;
    xStep=1;
   }
   if(this.xSubpixels<=-this.speedDenominator) {
    this.xSubpixels+=this.speedDenominator;
    xStep-=1;
   }
   if(this.ySubpixels>=this.speedDenominator) {
    this.ySubpixels-=this.speedDenominator;
    yStep=1;
   }
   if(this.ySubpixels<=-this.speedDenominator) {
    this.ySubpixels+=this.speedDenominator;
    yStep=-1;
   }
   if(xStep || yStep) {
    this.runPixelStep(xStep,yStep,collisionStyle);    
   }
  }
 }
 runPixelStep(xStep, yStep, collisionStyle) {
  var destLeft=this.left+xStep, destTop=this.top+yStep;
  if(collisionStyle=="warp") {
   if(this.wouldOverlapFullSolid(destLeft,destTop)) {
    // inside wall: back off
    this.xSubpixels=0;
    this.ySubpixels=0;
    this.performWarp();
    return;
   }
  }
  else {
   if(this.wouldOverlapAnySolid(destLeft,destTop)) {
    // hitting something: turn around, or enter despawning animation,
    // or despawn instantly
    this.xSubpixels=0;
    this.ySubpixels=0;
    if(collisionStyle=="patrol") {
     this.performPatrolBounce();
    }
    else { 
     this.enterDespawnState();
    }
    return;
   }
  }
  // passed the overlap checks, so actually move the step
  this.left+=xStep;
  this.top+=yStep;
 }
 wouldOverlapFullSolid(destLeft,destTop) {
  for(var y=destTop;y<destTop+this.height;++y) {
   for(var x=destLeft; x<destLeft+this.width;++x) {
    if(!this.isTerrainPixelSolid(x,y)) { return false; }
   }
  }
  return true;
 }
 wouldOverlapAnySolid(destLeft,destTop) {
  for(var y=destTop;y<destTop+this.height;++y) {
   for(var x=destLeft; x<destLeft+this.width;++x) {
    if(this.isTerrainPixelSolid(x,y)) { return true; }
   }
  }
  return false
 }
 isTerrainPixelSolid(x,y) {
  if(y<0 || y>=currentLevel.height || x<0 || x>=currentLevel.width) {
   return true;
  }
  var pixelIndex=y*currentLevel.width+x;
  return currentLevel.pixelData[pixelIndex*4+3]>0;
 }
 performWarp() {
  if(this.memoizedWarpResult) {
   // avoid doing the scan repeatedly
   this.left=this.memoizedWarpResult.left;
   this.top=this.memoizedWarpResult.top;
   return;
  }
  var xDir=0, yDir=0;
  if(this.xVelocityNumerator>0) { xDir=-1; }
  if(this.xVelocityNumerator<0) { xDir=1; }
  if(this.yVelocityNumerator>0) { yDir=-1; }
  if(this.yVelocityNumerator<0) { yDir=1; }
  if(!xDir && !yDir) {
   // this loop would become infinite..
   return;
  }
  var destLeft=this.left+xDir, destTop=this.top+yDir;
  while(!this.wouldOverlapFullSolid(destLeft+xDir,destTop+yDir)) {
   destLeft+=xDir; destTop+=yDir;
  }
  this.memoizedWarpResult={
   left:destLeft,
   top:destTop
  }
  this.left=destLeft;
  this.top=destTop;
 }
 performPatrolBounce() {
  this.xVelocityNumerator=-this.xVelocityNumerator;
  this.yVelocityNumerator=-this.yVelocityNumerator;
  if(this.altSpritesheetTop!=undefined) {
   var tmp=this.altSpritesheetTop;
   this.altSpritesheetTop=this.spritesheetTop;
   this.spritesheetTop=tmp;
   this.animationCycleCounter=0;
   this.texLeft=this.spritesheetLeft;
   this.texTop=this.spritesheetTop;
  }
 }
 runXSineMovement() {
  if(this.xSineGameFramesPerLoop<=0) { return; }
  // floating-point intermediates are used here; the result might
  // transiently vary by one pixel of distance depending on implementations.
  this.xSinePhaseCycle=(this.xSinePhaseCycle+1)%(this.xSineGameFramesPerLoop);
  var initialOffset=
      this.xSineRange*Math.sin(this.xSineInitialPhase*Math.PI/180);
  var zeroPosition=this.xSineInitialLeft-initialOffset;
  var cyclePosition=
      (Math.PI*2*this.xSinePhaseCycle/this.xSineGameFramesPerLoop) +
      (this.xSineInitialPhase*Math.PI/180);
  var left=zeroPosition+this.xSineRange*Math.sin(cyclePosition)
  this.left=Math.floor(left+.5);
 }
 runYSineMovement() {
  if(this.ySineGameFramesPerLoop<=0) { return; }
  // floating-point intermediates are used here; the result might
  // transiently vary by one pixel of distance depending on implementations.
  this.ySinePhaseCycle=(this.ySinePhaseCycle+1)%(this.ySineGameFramesPerLoop);
  var initialOffset=
      this.ySineRange*Math.sin(this.ySineInitialPhase*Math.PI/180);
  var zeroPosition=this.ySineInitialTop-initialOffset;
  var cyclePosition=
      (Math.PI*2*this.ySinePhaseCycle/this.ySineGameFramesPerLoop) +
      (this.ySineInitialPhase*Math.PI/180);
  var top=zeroPosition+this.ySineRange*Math.sin(cyclePosition);
  this.top=Math.floor(top+.5);
 }
 
 enterDespawnState() {
  if(this.despawnAnimationFrameCount) {
   this.spritesheetLeft=this.despawnSpritesheetLeft;
   this.spritesheetTop=this.despawnSpritesheetTop;
   this.spritesheetGrid=this.despawnSpritesheetGrid;
   this.spritesheetTop=this.despawnSpritesheetTop;
   this.animationFrameCount=this.despawnAnimationFrameCount|0;      
   this.gameFramesPerAnimationFrame=this.gameFramesPerDespawnAnimationFrame;
   this.texLeft=this.spritesheetLeft;
   this.texTop=this.spritesheetTop;
   this.animationCycleCounter=0;
   this.moveType="despawn";
  }
  else {
   this.despawned=true;
  }
 }
 handlePeriodicSpawns() {
  for(var i in this.periodicSpawns) {
   var spawnEntry=this.periodicSpawns[i];
   if(spawnEntry.intervalFrames>0) {
    spawnEntry.offsetFrames=
     (spawnEntry.offsetFrames-1)%spawnEntry.intervalFrames;
    if(spawnEntry.offsetFrames==0) {
     var ctor=spawnEntry.spriteClass;
     currentLevel.sprites.push(new ctor(this.left,this.top));
    }
   }
  }
 }		     
 visibleForCamera(cameraLeft, cameraTop) {
  if(this.blinkFrames && this.blinkFrames%8>0 &&
     this.blinkFrames%8<5 && !this.blinkSuppressed) {
   return false;
  }
  return this.left+this.width>cameraLeft &&
   this.top+this.height>cameraTop &&
   this.left<cameraLeft+VIEWPORT_WIDTH &&
   this.top<cameraTop+VIEWPORT_HEIGHT &&
   this.texLeft!=undefined;
 }
 drawForCamera(cameraLeft,cameraTop) {
  if(this.texLeft!=undefined) {
   gameContext.drawImage(spritesheet,
			 this.texLeft, this.texTop,
			 this.width, this.height,
			 this.left-cameraLeft, this.top-cameraTop,
			 this.width, this.height);
  }
 }
 testPlayerCollision(x,y) {
  if(x<this.left || y<this.top ||
     x>=this.left+this.width || y>=this.top+this.height ||
     this.texLeft==undefined) {
   return false;
  }
  var texX=x-this.left+this.texLeft;
  var texY=y-this.top+this.texTop;
  if(texX<0 || texY<0 ||
     texX>=SPRITESHEET_MASK.width || texY>=SPRITESHEET_MASK.height) {
   return false;
  }
  var texIndex=texY*SPRITESHEET_MASK.width+texX;
  return SPRITESHEET_MASK.alpha[texIndex]>0;
 }
 onPlayerTouch() {
  if(currentLevel.player.invulnerabilityFrames>0 || this.blinkFrames>0) {
   return false; 
  }
  currentLevel.player.takeDamage(true);
  this.blinkFrames=DAMAGE_INVULNERABILITY_FRAMES;
  return true;
 }
 onPlayerReachedCheckpoint() {
  if(this.moveType=="projectile" || this.moveType=="despawn") {
   this.despawned=true;
  }
 }
}

class BigFractalSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   36,36 // width,height
  );
  this.setAnimation(
   0,1024, // left,top in spritesheet
   48, 38, // spritesheet grid size, number of grid spaces
   3 // animation timing
  );
 }
 static stamp=[
  ".................##.................",
  "................####................",
  "................####................",
  "...............######...............",
  "..............########..............",
  ".............##########.............",
  ".............##########.............",
  ".............##########.............",
  ".............##########.............",
  "..........#...########...#..........",
  ".........####..######..####.........",
  "..........####.######.####..........",
  "..........################..........",
  ".....####..##############..####.....",
  "....######..############..######....",
  "...##############################...",
  ".##################################.",
  "####################################",
  "####################################",
  ".##################################.",
  "...##############################...",
  "....######..############..######....",
  ".....####..##############..####.....",
  "..........################..........",
  "..........####.######.####..........",
  ".........####..######..####.........",
  "..........#...########...#..........",
  ".............##########.............",
  ".............##########.............",
  ".............##########.............",
  ".............##########.............",
  "..............########..............",
  "...............######...............",
  "................####................",
  "................####................",
  ".................##................."
 ]
}

registerSprite(BigFractalSprite);


class GreenFractalSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   36,36 // width,height
  );
  this.setAnimation(
   0,1072, // left,top in spritesheet
   48, 38, // spritesheet grid size, number of grid spaces
   3 // animation timing
  );
 }
 static stamp=[
  ".................##.................",
  "................#..#................",
  "................#..#................",
  "...............#....#...............",
  "..............#......#..............",
  ".............#........#.............",
  ".............#........#.............",
  ".............#........#.............",
  ".............#........#.............",
  "..........#...#......#...#..........",
  ".........#.##..#....#..##.#.........",
  "..........#..#.#....#.#..#..........",
  "..........#...#......#...#..........",
  ".....####..#............#..####.....",
  "....#....#..#..........#..#....#....",
  "...#......###..........###......#...",
  ".##..............................##.",
  "#..................................#",
  "#..................................#",
  ".##..............................##.",
  "...#......###..........###......#...",
  "....#....#..#..........#..#....#....",
  ".....####..#............#..####.....",
  "..........#...#......#...#..........",
  "..........#..#.#....#.#..#..........",
  ".........#.##..#....#..##.#.........",
  "..........#...#......#...#..........",
  ".............#........#.............",
  ".............#........#.............",
  ".............#........#.............",
  ".............#........#.............",
  "..............#......#..............",
  "...............#....#...............",
  "................#..#................",
  "................#..#................",
  ".................##................."
 ]
};

registerSprite(GreenFractalSprite);


class UpwardTentacleSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   23,31 // width,height
  );
  this.setAnimation(
   0,1120, // left,top in spritesheet
   24, 23, // spritesheet grid size, number of grid spaces
   3 // animation timing
  );
 }
 static stamp=[
  "..........##...........",
  "..........###..........",
  "..........###..........",
  "..........####.....###.",
  "..###.....####...######",
  ".#####.....###..#######",
  ".#####.....###.########",
  ".#######...###########.",
  "..######...#########...",
  "...######..#######.....",
  "....#####...#####......",
  "....######.#####.......",
  "....###########........",
  ".....##########........",
  ".....##########........",
  ".....##########........",
  ".....###########.......",
  "......##########.......",
  "......##########.......",
  ".....###########.......",
  ".....###########.......",
  "....###########........",
  "....###########........",
  "....##########.........",
  "....#########..........",
  ".....########..........",
  "......######...........",
  ".......#####...........",
  ".......#####...........",
  "........####...........",
  "........####..........."
 ]
};

registerSprite(UpwardTentacleSprite);

class DownwardTentacleSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   23,31 // width,height
  );
  this.setAnimation(
   0,1152, // left,top in spritesheet
   24, 23, // spritesheet grid size, number of grid spaces
   5 // animation timing
  );
 }
 static stamp=[
  "...........####........",
  "...........####........",
  "...........#####.......",
  "...........#####.......",
  "...........######......",
  "..........########.....",
  "..........#########....",
  ".........##########....",
  "........###########....",
  "........###########....",
  ".......###########.....",
  ".......###########.....",
  ".......##########......",
  ".......##########......",
  ".......###########.....",
  "........##########.....",
  "........##########.....",
  "........##########.....",
  "........###########....",
  ".......#####.######....",
  "......#####...#####....",
  ".....#######..######...",
  "...#########...######..",
  ".###########...#######.",
  "########.###.....#####.",
  "#######..###.....#####.",
  "######...####.....###..",
  ".###.....####..........",
  "..........###..........",
  "..........###..........",
  "...........##.........."
 ]
}

registerSprite(DownwardTentacleSprite);

class RightwardTentacleSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   31,23 // width,height
  );
  this.setAnimation(
   0,1184, // left,top in spritesheet
   32, 23, // spritesheet grid size, number of grid spaces
   4 // animation timing
  );
 }
 static stamp=[
  "...............................",
  ".......................###.....",
  "......................#####....",
  ".....................######....",
  "......####........#########....",
  ".....#######..############.....",
  "....####################.......",
  "..######################.......",
  "######################.........",
  "####################...........",
  "###################.......#####",
  "####################.##########",
  ".....#########################.",
  ".......#####################...",
  "........################.......",
  "..........#####....######......",
  "....................######.....",
  ".....................######....",
  "......................#####....",
  "......................######...",
  ".......................#####...",
  ".......................#####...",
  "........................###...."
 ]
}

registerSprite(RightwardTentacleSprite);

class LeftwardTentacleSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   31,23 // width,height
  );
  this.setAnimation(
   0,1216, // left,top in spritesheet
   32, 23, // spritesheet grid size, number of grid spaces
   3 // animation timing
  );
 }
 static stamp=[
  "....###........................",
  "...#####.......................",
  "...#####.......................",
  "...######......................",
  "....#####......................",
  "....######.....................",
  ".....######....................",
  "......######....#####..........",
  ".......################........",
  "...#####################.......",
  ".#########################.....",
  "##########.####################",
  "#####.......###################",
  "...........####################",
  ".........######################",
  ".......######################..",
  ".......####################....",
  ".....############..#######.....",
  "....#########........####......",
  "....######.....................",
  "....#####......................",
  ".....###.......................",
  "..............................."
 ]
}

registerSprite(LeftwardTentacleSprite);

class SnailSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   13,8 // width,height
  );
  this.setAnimation(
   0,1280, // left,top in spritesheet
   16, 4, // spritesheet grid size, number of grid spaces
   8 // animation timing
  );
  this.setPatrolMove(
   1,0, // x,y velocity in subpixels
   8, // number of subpixels per pixel
   1296 // top in spritesheet for reverse direction
  );
 }
 static stamp=[
  "..#####......",
  ".#######.....",
  "#########....",
  "#########....",
  "#########..##",
  "#########.##.",
  "#############",
  ".##########.."
 ]
}

registerSprite(SnailSprite);

class GreenSnailSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   13,8 // width,height
  );
  this.setAnimation(
   0,1328, // left,top in spritesheet
   16, 4, // spritesheet grid size, number of grid spaces
   6 // animation timing
  );
  this.setPatrolMove(
   -1,0, // x,y velocity in subpixels
   6, // number of subpixels per pixel
   1312 // top in spritesheet for reverse direction
  );
 }
 static stamp=[
  "......#####..",
  ".....#.....#.",
  "....#..###..#",
  "....#.#...#.#",
  "##..#.#.#.#.#",
  ".##.#.#..#..#",
  "#####..#...##",
  "..##########."
 ]
}
registerSprite(GreenSnailSprite);

class UpRightDiagonalSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   8,8 // width,height
  );
  this.setAnimation(
   0,1360, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   1 // animation timing
  );
  this.setPatrolMove(
   1,-1, // x,y velocity in subpixels
   2 // number of subpixels per pixel
   // omit fourth argument, animation doesn't change on turnaround
  );
 }
 static stamp=[
  "....####",
  "...#####",
  "..######",
  ".#######",
  "#######.",
  "######..",
  "#####...",
  "####....",
 ]
}

registerSprite(UpRightDiagonalSprite);



class DownRightDiagonalSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   8,8 // width,height
  );
  this.setAnimation(
   0,1344, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   1 // animation timing
  );
  this.setPatrolMove(
   1,1, // x,y velocity in subpixels
   2, // number of subpixels per pixel
   // omit fourth argument, animation doesn't change on turnaround
  );
 }
 static stamp=[
  "####....",
  "#####...",
  "######..",
  "#######.",
  ".#######",
  "..######",
  "...#####",
  "....####",
 ]
}

registerSprite(DownRightDiagonalSprite);

class CubePillarSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   8,27 // width,height
  );
  this.setAnimation(
   0,1408, // left,top in spritesheet
   16, 4, // spritesheet grid size, number of grid spaces
   6 // animation timing
  );
  this.setPatrolMove(
   0,1, // x,y velocity in subpixels
   6, // number of subpixels per pixel
   // omit fourth argument, animation doesn't change on turnaround
  );
 }
 static stamp=[
  ".######.",
  ".######.",
  ".######.",
  ".######.",
  ".######.",
  "########",
  "########",
  "########",
  "########",
  "########",
  ".######.",
  ".######.",
  ".######.",
  "########",
  "########",
  "########",
  "########",
  "########",
  ".######.",
  ".######.",
  ".######.",
  "########",
  "########",
  "########",
  "########",
  "########",
  "..####..",
 ]
}

registerSprite(CubePillarSprite);

class SpinnySquareSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   16,9 // width,height
  );
  this.setAnimation(
   0,1376, // left,top in spritesheet
   32, 8, // spritesheet grid size, number of grid spaces
   5 // animation timing
  );
  this.setPatrolMove(
   1,0, // x,y velocity in subpixels
   3, // number of subpixels per pixel
   1392 // top in spritesheet for reverse direction
  );
 }
 static stamp=[
  "......####......",
  "....########....",
  "..####....####..",
  "#####......#####",
  "######....######",
  "################",
  "..############..",
  "....########....",
  "......####......",  
 ];
}

registerSprite(SpinnySquareSprite);

class RedBugSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   8,5 // width,height
  );
  this.setAnimation(
   0,1440, // left,top in spritesheet
   16, 4, // spritesheet grid size, number of grid spaces
   4 // animation timing
  );
  this.setPatrolMove(
   0,-1, // x,y velocity in subpixels
   2, // number of subpixels per pixel
   // omit fourth argument, animation doesn't change on turnaround
  );
 }
 static stamp=[
  ".#####..",
  "########",
  "########",
  "######..",
  ".###....",
 ]
}

registerSprite(RedBugSprite);

class ChipSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   15,8 // width,height
  );
  this.setAnimation(
   0,1472, // left,top in spritesheet
   16, 6, // spritesheet grid size, number of grid spaces
   4 // animation timing
  );
  this.setPatrolMove(
   -1,0, // x,y velocity in subpixels
   8, // number of subpixels per pixel
   1456 // top in spritesheet for reverse direction
  );
 }
 static stamp=[
  "###############",
  "#.............#",
  "##............#",
  "#.............#",
  "#.............#",
  "###############",
  "..#.#.#.#.#.#..",
  "..#.#.#.#.#.#..",
 ] 
}

registerSprite(ChipSprite);

class MetalWarpingSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   9,9 // width,height
  );
  this.setAnimation(
   0,1488, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   0 // animation timing
  );
  this.setXWarpMove(
   -1, // x velocity in subpixels
   2, // number of subpixels per pixel
  );
  this.setLayer(-1); // draw behind terrain
 }
 static stamp=[
  "##.....##",
  "###.#.###",
  ".#######.",
  "..##.##..",
  ".##.#.##.",
  "..##.##..",
  ".#######.",
  "###.#.###",
  "##.....##"
 ] 
}

registerSprite(MetalWarpingSprite);


class GreenWarpingSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   9,9 // width,height
  );
  this.setAnimation(
   0,1504, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   0 // animation timing
  );
  this.setXWarpMove(
   2, // x velocity in subpixels
   3, // number of subpixels per pixel
  );
  this.setLayer(-1); // draw behind terrain
 }
 static stamp=[
  "##.....##",
  "#.#.#.#.#",
  ".#.#.#.#.",
  "..#...#..",
  ".#.....#.",
  "..#...#..",
  ".#.#.#.#.",
  "#.#.#.#.#",
  "##.....##"
 ] 
}

registerSprite(GreenWarpingSprite);



class BatSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   6,6 // width,height
  );
  this.setAnimation(
   0,1520, // left,top in spritesheet
   8, 6, // spritesheet grid size, number of grid spaces
   3 // animation timing
  );
  this.setXSinusoidMove(
   16, // range of movement
   0, // initial sine phase in degrees (start X at center and increasing)
   256, // frames per loop
  );
  this.setYSinusoidMove(
   16, // range of movement
   -90, // initial sine phase in degrees (start Y at minimum)
   256, // frames per loop
  );  
 }
 static stamp=[
  "##..##",
  "#.##.#",
  "#....#",
  ".#..#.",
  "..##.."
 ] 
}

registerSprite(BatSprite);


class GlowBatSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   16,14 // width,height
  );
  this.setAnimation(
   0,1536, // left,top in spritesheet
   32, 6, // spritesheet grid size, number of grid spaces
   3 // animation timing
  );
  this.setXSinusoidMove(
   16, // range of movement
   180, // initial sine phase in degrees (start X at center and decreasing)
   205, // frames per loop
  );
  this.setYSinusoidMove(
   16, // range of movement
   -90, // initial sine phase in degrees (start Y at minimum)
   205, // frames per loop
  );  
 }
 static stamp=[
  "...###....###...",
  "..#...####...#..",
  ".#.##########.#.",
  ".#.#........#.#.",
  ".#.#........#.#.",
  ".#.#..####..#.#.",
  "#.#...#..#...#.#",
  "#.#..##..##..#.#",
  "#.#..#.##.#..#.#",
  ".#.#.##..##.#.#.",
  "..#.#..##..#.#..",
  "...#.##..##.#...",
  "....#..##..#....",
  ".....##..##....."
 ] 
}

registerSprite(GlowBatSprite);

class HorizontalOrganSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   20,20 // width,height
  );
  this.setAnimation(
   0,1568, // left,top in spritesheet
   32, 6, // spritesheet grid size, number of grid spaces
   5 // animation timing
  );
  this.setXSinusoidMove(
   17, // range of movement
   0, // initial sine phase in degrees (start X at center and increasing)
   205 // frames per loop
  );
 }
 static stamp=
  [
   ".....##########.....",
   "....#..........##...",
   "..##.............#..",
   ".#................#.",
   "#..................#",
   "#..................#",
   "#....#........#....#",
   "#...#..........#...#",
   "#..#............#..#",
   "#.################.#",
   "#..#...........#...#",
   "#...#.........#...#.",
   ".#...#.......#...#..",
   ".#...............#..",
   "..#.............#...",
   "..#.............#...",
   "...#............#...",
   "...#...........#....",
   "....#..........#....",
   ".....##########....."
  ]
}

registerSprite(HorizontalOrganSprite);

class VerticalOrganSprite extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   20,20 // width,height
  );
  this.setAnimation(
   0,1568, // left,top in spritesheet
   32, 6, // spritesheet grid size, number of grid spaces
   7 // animation timing
  );
  this.setYSinusoidMove(
   8, // range of movement
   0, // initial sine phase in degrees (start Y at center and increasing)
   114 // frames per loop
  );
 }
 static stamp=[
  ".....##########.....",
  "....#..........##...",
  "..##.....#.......#..",
  ".#......###.......#.",
  "#......#.#.#.......#",
  "#.....#..#..#......#",
  "#........#.........#",
  "#........#.........#",
  "#........#.........#",
  "#........#.........#",
  "#........#.........#",
  "#........#........#.",
  ".#.......#.......#..",
  ".#.......#.......#..",
  "..#...#..#..#...#...",
  "..#....#.#.#....#...",
  "...#....###.....#...",
  "...#.....#.....#....",
  "....#..........#....",
  ".....##########....."
 ]
}

registerSprite(VerticalOrganSprite);
 
class DropletSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   10,7 // width,height
  );
  this.setAnimation(
   0,1600, // left,top in spritesheet
   16, 21, // spritesheet grid size, number of grid spaces
   3 // animation timing
  );
  this.addPeriodicSpawn(
   DropletSpawn, // sprite class to spawn
   21*3, // 21 sprite frames by 3 timing frames, so once per animation loop
   0 // no timing offset
  )
 }
 static stamp=
  [
   "##########",
   "...####...",
   "....##....",
   "....##....",
   "...#.##...",
   "...####...",
   "....##....",
  ]
}

registerSprite(DropletSpawner);

class DropletSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x+1, y+2, // left,top are offset by 1,2 from spawner's left,top position
   8,5 // width,height
  )
  this.setAnimation(
   0,1616, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setDespawnAnimation(
   16,1616, // left,top in spritesheet
   16, 7, // spritesheet grid size, number of grid spaces
   2 // animation timing
  )
  this.setProjectileMove(
   0,1, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   1, // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.
}

class GreenDropletSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   10,7 // width,height
  );
  this.setAnimation(
   0,1632, // left,top in spritesheet
   16, 21, // spritesheet grid size, number of grid spaces
   2 // animation timing
  );
  this.addPeriodicSpawn(
   GreenDropletSpawn, // sprite class to spawn
   21*2, // 21 sprite frames by 2 timing frames, so once per animation loop
   0 // no timing offset
  )
 }
 static stamp=
  [
   "###....###",
   "...#..#...",
   "....##....",
   "....##....",
   "...#..#...",
   "...#..#...",
   "....##....",
  ]
}

registerSprite(GreenDropletSpawner);

class GreenDropletSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x+1, y+2, // left,top are offset by 1,2 from spawner's left,top position
   8,5 // width,height
  )
  this.setAnimation(
   0,1648, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setDespawnAnimation(
   16,1648, // left,top in spritesheet
   16, 7, // spritesheet grid size, number of grid spaces
   2 // animation timing
  )
  this.setProjectileMove(
   0,1, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   1, // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.
}

class SteamSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   4,8 // width,height
  );
  this.setAnimation(
   0,1664, // left,top in spritesheet
   16, 52, // spritesheet grid size, number of grid spaces
   2 // animation timing
  );
  this.addPeriodicSpawn(
   SteamSpawn, // sprite class to spawn
   52*2, // 52 sprite frames by 2 timing frames, so once per animation loop
   0 // no timing offset
  )
  this.suppressBlink(); // don't flicker if player crashes into this
 }
 static stamp=[
  ".##.",
  "##.#",
  "#.##",
  "##.#",
  ".##."
 ]
}

registerSprite(SteamSpawner);

class SteamSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x, y, // left,top equal spawner's left,top position
   4,5 // width,height
  )
  this.setAnimation(
   0,1680, // left,top in spritesheet
   16, 12, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setDespawnAnimation(
   192,1680, // left,top in spritesheet
   16, 3, // spritesheet grid size, number of grid spaces
   3 // animation timing
  )
  this.setProjectileMove(
   0,-1, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   4, // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.
}

class CubeSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   8,10 // width,height
  );
  this.setAnimation(
   0,1696, // left,top in spritesheet
   16, 56, // spritesheet grid size, number of grid spaces
   3 // animation timing
  );
  this.addPeriodicSpawn(
   CubeSpawnUp, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   25*3 // fire on 25th frame of animation loop
  )
  this.addPeriodicSpawn(
   CubeSpawnLeft, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   25*3 // fire on 25th frame of animation loop
  )
  this.addPeriodicSpawn(
   CubeSpawnRight, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   25*3 // fire on 25th frame of animation loop
  )
  this.addPeriodicSpawn(
   CubeSpawnDown, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   25*3 // fire on 25th frame of animation loop
  )
  this.addPeriodicSpawn(
   CubeSpawnUpLeft, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   53*3 // fire on 53rd frame of animation loop
  )
  this.addPeriodicSpawn(
   CubeSpawnUpRight, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   53*3 // fire on 53rd frame of animation loop
  )
  this.addPeriodicSpawn(
   CubeSpawnDownLeft, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   53*3 // fire on 53rd frame of animation loop
  )
  this.addPeriodicSpawn(
   CubeSpawnDownRight, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   53*3 // fire on 53rd frame of animation loop
  )
 }
 static stamp=
  [
   "..####..",
   "###..###",
   "#......#",
   "#......#",
   "#......#",
   "###..###",
   "..####..",
  ]
}

registerSprite(CubeSpawner);

/* There are eight kinds of cube spawn that differ only in velocity
   and spawn position, so a base class makes them more concise to write. */
class AbstractCubeSpawn extends CustomSprite {
 constructor(x,y,xv,yv) {
  super();
  this.setRect(
   x, y, // left,top as defined in subclass superconstructor call
   6,5 // width,height
  )
  this.setAnimation(
   0,1712, // left,top in spritesheet
   16, 3, // spritesheet grid size, number of grid spaces
   8 // animation timing
  )
  this.setDespawnAnimation(
   48,1712, // left,top in spritesheet
   16, 2, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   xv,yv, // x,y velocity in subpixels as defined in subclass's super call
   0,0, // x,y acceleration in subpixels
   6, // number of subpixels per pixel
  )
 }
}
// Each shot has an offset from the spawner and a velocity.
class CubeSpawnUp extends AbstractCubeSpawn {
 constructor(x,y) { super(x+1,y+1,0,-3); }
}
class CubeSpawnLeft extends AbstractCubeSpawn {
 constructor(x,y) { super(x,y+2,-3,0); }
}
class CubeSpawnRight extends AbstractCubeSpawn {
 constructor(x,y) { super(x+2,y+2,3,0); }
}
class CubeSpawnDown extends AbstractCubeSpawn {
 constructor(x,y) { super(x+1,y+3,0,3); }
}
class CubeSpawnUpLeft extends AbstractCubeSpawn {
 constructor(x,y) { super(x,y+1,-2,-2); }
}
class CubeSpawnUpRight extends AbstractCubeSpawn {
 constructor(x,y) { super(x+2,y+1,2,-2); }
}
class CubeSpawnDownLeft extends AbstractCubeSpawn {
 constructor(x,y) { super(x,y+3,-2,2); }
}
class CubeSpawnDownRight extends AbstractCubeSpawn {
 constructor(x,y) { super(x+2,y+3,2,2); }
}

class GreenCubeSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   8,10 // width,height
  );
  this.setAnimation(
   0,1728, // left,top in spritesheet
   16, 56, // spritesheet grid size, number of grid spaces
   3 // animation timing
  );
  this.addPeriodicSpawn(
   GreenCubeSpawnUpLeft, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   25*3 // fire on 25th frame of animation loop
  )
  this.addPeriodicSpawn(
   GreenCubeSpawnUpRight, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   25*3 // fire on 25th frame of animation loop
  )
  this.addPeriodicSpawn(
   GreenCubeSpawnDownLeft, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   25*3 // fire on 25th frame of animation loop
  )
  this.addPeriodicSpawn(
   GreenCubeSpawnDownRight, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   25*3 // fire on 25th frame of animation loop
  )
  this.addPeriodicSpawn(
   GreenCubeSpawnUp, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   53*3 // fire on 53rd frame of animation loop
  )
  this.addPeriodicSpawn(
   GreenCubeSpawnLeft, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   53*3 // fire on 53rd frame of animation loop
  )
  this.addPeriodicSpawn(
   GreenCubeSpawnDown, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   53*3 // fire on 53rd frame of animation loop
  )
  this.addPeriodicSpawn(
   GreenCubeSpawnRight, // sprite class to spawn
   56*3, // 56 sprite frames by 3 timing frames, so once per animation loop
   53*3 // fire on 53rd frame of animation loop
  )
 }
 static stamp=
  [
   "..####..",
   "##....##",
   "#.##..##",
   "#...##.#",
   "#...#..#",
   "##..#.##",
   "..####..",
  ]
}

registerSprite(GreenCubeSpawner);

/* As with the other cube, there are eight kinds of cube spawn that
   differ only in velocity and spawn position, so a base class can hold
   the common values */
class AbstractGreenCubeSpawn extends CustomSprite {
 constructor(x,y,xv,yv) {
  super();
  this.setRect(
   x, y, // left,top as defined in subclass superconstructor call
   8,7 // width,height
  )
  this.setAnimation(
   0,1744, // left,top in spritesheet
   16, 3, // spritesheet grid size, number of grid spaces
   8 // animation timing
  )
  this.setDespawnAnimation(
   48,1744, // left,top in spritesheet
   16, 2, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   xv,yv, // x,y velocity in subpixels as defined in subclass's super call
   0,0, // x,y acceleration in subpixels
   6, // number of subpixels per pixel
  )
 }
}
// Each shot has an offset from the spawner and a velocity.
class GreenCubeSpawnUp extends AbstractGreenCubeSpawn {
 constructor(x,y) { super(x,y,0,-3); }
}
class GreenCubeSpawnLeft extends AbstractGreenCubeSpawn {
 constructor(x,y) { super(x-1,y+1,-3,0); }
}
class GreenCubeSpawnRight extends AbstractGreenCubeSpawn {
 constructor(x,y) { super(x+1,y+1,3,0); }
}
class GreenCubeSpawnDown extends AbstractGreenCubeSpawn {
 constructor(x,y) { super(x,y+2,0,3); }
}
class GreenCubeSpawnUpLeft extends AbstractGreenCubeSpawn {
 constructor(x,y) { super(x-1,y,-2,-2); }
}
class GreenCubeSpawnUpRight extends AbstractGreenCubeSpawn {
 constructor(x,y) { super(x+1,y,2,-2); }
}
class GreenCubeSpawnDownLeft extends AbstractGreenCubeSpawn {
 constructor(x,y) { super(x-1,y+2,-2,2); }
}
class GreenCubeSpawnDownRight extends AbstractGreenCubeSpawn {
 constructor(x,y) { super(x+1,y+2,2,2); }
}

class ArcSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   11,11 // width,height
  );
  this.setAnimation(
   0,1760, // left,top in spritesheet
   16, 12, // spritesheet grid size, number of grid spaces
   2 // animation timing
  );
  this.addPeriodicSpawn(
   ArcSpawn1A, // sprite class to spawn
   12*2*24, // firing pattern repeats over 24 animation loops
   0, // fire at repeat point of pattern
  )
  this.addPeriodicSpawn(
   ArcSpawn1B, // sprite class to spawn
   12*2*24, // firing pattern repeats over 24 animation loops
   0, // fire at repeat point of pattern
  )
  this.addPeriodicSpawn(
   ArcSpawn2A, // sprite class to spawn
   12*2*24, // firing pattern repeats over 24 animation loops
   12*2*3, // fire after third loop in pattern
  )
  this.addPeriodicSpawn(
   ArcSpawn2B, // sprite class to spawn
   12*2*24, // firing pattern repeats over 24 animation loops
   12*2*3, // fire after third loop in pattern
  )
  this.addPeriodicSpawn(
   ArcSpawn3A, // sprite class to spawn
   12*2*24, // firing pattern repeats over 24 animation loops
   12*2*6, // fire after third loop in pattern
  )
  this.addPeriodicSpawn(
   ArcSpawn3B, // sprite class to spawn
   12*2*24, // firing pattern repeats over 24 animation loops
   12*2*6, // fire after third loop in pattern
  )
  this.addPeriodicSpawn(ArcSpawn4A, 12*2*24, 12*2*9);
  this.addPeriodicSpawn(ArcSpawn4B, 12*2*24, 12*2*9);
  this.addPeriodicSpawn(ArcSpawn5A, 12*2*24, 12*2*12);
  this.addPeriodicSpawn(ArcSpawn5B, 12*2*24, 12*2*12);
  this.addPeriodicSpawn(ArcSpawn6A, 12*2*24, 12*2*15);
  this.addPeriodicSpawn(ArcSpawn6B, 12*2*24, 12*2*15);
  this.addPeriodicSpawn(ArcSpawn7A, 12*2*24, 12*2*18);
  this.addPeriodicSpawn(ArcSpawn7B, 12*2*24, 12*2*18);
  this.addPeriodicSpawn(ArcSpawn8A, 12*2*24, 12*2*21);
  this.addPeriodicSpawn(ArcSpawn8B, 12*2*24, 12*2*21);
 }
 static stamp=
  [
   "....###....",
   "..######...",
   ".########..",
   "###...###..",
   "##..####...",
   "...#####...",
   "...####..##",
   "..###...###",
   "..########.",
   "...######..",
   "....###...."
  ]
}

registerSprite(ArcSpawner);

/* Arc spawns vary in just a few parameters; a base class handles the rest */
class AbstractArcSpawn extends CustomSprite {
 constructor(x,y,xv,yv,xa,ya) {
  super();
  this.setRect(
   x, y, // left,top as defined in subclass superconstructor call
   8,7 // width,height
  )
  this.setAnimation(
   0,1776, // left,top in spritesheet
   8, 16, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setDespawnAnimation(
   0,1888, // left,top in spritesheet
   8, 9, // spritesheet grid size, number of grid spaces
   2 // animation timing
  )
  this.setProjectileMove(
   xv,yv, // x,y velocity in subpixels as defined in subclass's super call
   xa,ya, // x,y acceleration in subpixels per sublass
   256, // number of subpixels per pixel
  )
 }
}
// Each shot has an offset from the spawner, a velocity, and an acceleration
class ArcSpawn1A extends AbstractArcSpawn {
 constructor(x,y) {
  super(x,y+2, -128,0, 2,2);
 }
}
class ArcSpawn1B extends AbstractArcSpawn {
 constructor(x,y) {
  super(x+3,y+1, 91,-91, -4,0);
 }
}
class ArcSpawn2A extends AbstractArcSpawn {
 constructor(x,y) {
  super(x+1,y+1, -91,-91, 0,3);
 }
}
class ArcSpawn2B extends AbstractArcSpawn { constructor(x,y) {
 super(x+4,y+2, 128,0, -3,-3);
} }
class ArcSpawn3A extends AbstractArcSpawn { constructor(x,y) {
 super(x+2,y, 0,-128, -2,2);
} }
class ArcSpawn3B extends AbstractArcSpawn { constructor(x,y) {
 super(x+3,y+3, 91,91, 0,-4);
} }
class ArcSpawn4A extends AbstractArcSpawn { constructor(x,y) {
 super(x+3,y+1, 91,-91, -3,0);
} }
class ArcSpawn4B extends AbstractArcSpawn { constructor(x,y) {
 super(x+2,y+4, 0,128, 3,-3);
} }
class ArcSpawn5A extends AbstractArcSpawn { constructor(x,y) {
 super(x+4,y+2, 128,0, -2,-2);
} }
class ArcSpawn5B extends AbstractArcSpawn { constructor(x,y) {
 super(x+1,y+3, -91,91, 4,0);
} }
class ArcSpawn6A extends AbstractArcSpawn { constructor(x,y) {
 super(x+3,y+3, 91,91, 0,-3);
} }
class ArcSpawn6B extends AbstractArcSpawn { constructor(x,y) {
 super(x,y+2, -128,0, 3,3);
} }
class ArcSpawn7A extends AbstractArcSpawn { constructor(x,y) {
 super(x+2,y+4, 0,128, 2,-2);
} }
class ArcSpawn7B extends AbstractArcSpawn { constructor(x,y) {
 super(x+1,y+1, -91,-91, 0,4);
} }
class ArcSpawn8A extends AbstractArcSpawn { constructor(x,y) {
 super(x+1,y+3, -91,91, 3,0);
} }
class ArcSpawn8B extends AbstractArcSpawn { constructor(x,y) {
 super(x+2,y, 0,-128, -3,3);
} }

class DownwardSparkSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y-7, // left,top in world; top is 7 pixels above the top of the stamp
   8,13 // width,height
  );
  this.setAnimation(
   0,1792, // left,top in spritesheet
   16, 30, // spritesheet grid size, number of grid spaces
   4 // animation timing
  );
  this.addPeriodicSpawn(
   DownwardSparkSpawn, // sprite class to spawn
   30*4, // 30 sprite frames by 3 timing frames, so once per animation loop
   11*4, // fire at 11th sprite frame
  )
  this.addPeriodicSpawn(
   DownwardSparkSpawn, // sprite class to spawn
   30*4, // 30 sprite frames by 3 timing frames, so once per animation loop
   21*4, // fire at 21st sprite frame
  )
  this.addPeriodicSpawn(
   DownwardSparkSpawn, // sprite class to spawn
   30*4, // 30 sprite frames by 3 timing frames, so once per animation loop
   0, // fire at repeat point of loop
  )
  this.suppressBlink(); // don't flicker if player crashes into this
 }
 static stamp=
  [
   "...##...",
   "..#..#..",
   "..#..#..",   
   "...##...",
   "..#..#..",
   "..#..#..",   
   "..#..#..",
   "###..###",
   ".#....#.",
   "..#..#..",
   "...##...",

  ] 
}

registerSprite(DownwardSparkSpawner);

class DownwardSparkSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x+1, y+7, // left,top are offset by 1,7 from spawner's left,top position
   6,6 // width,height
  )
  this.setAnimation(
   0,1808, // left,top in spritesheet
   16, 16, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setDespawnAnimation(
   256,1808, // left,top in spritesheet
   16, 4, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   0,4, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   3, // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.  
}

class UpwardSparkSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y+5, // left,top in world; top is 5 pixels below the top of the stamp
   8,13 // width,height
  );
  this.setAnimation(
   0,1816, // left,top in spritesheet
   16, 30, // spritesheet grid size, number of grid spaces
   4 // animation timing
  );
  this.addPeriodicSpawn(
   UpwardSparkSpawn, // sprite class to spawn
   30*4, // 30 sprite frames by 3 timing frames, so once per animation loop
   6*4, // fire at 6th sprite frame
  )
  this.addPeriodicSpawn(
   UpwardSparkSpawn, // sprite class to spawn
   30*4, // 30 sprite frames by 3 timing frames, so once per animation loop
   16*4, // fire at 16st sprite frame
  )
  this.addPeriodicSpawn(
   UpwardSparkSpawn, // sprite class to spawn
   30*4, // 30 sprite frames by 3 timing frames, so once per animation loop
   25*4, // fire at 25th sprite frame
  )
  this.suppressBlink(); // don't flicker if player crashes into this
 }
 static stamp=
  [
   "...##...",
   "..#..#..",
   ".#....#.",   
   "###..###",
   "..#..#..",
   "..#..#..",   
   "..#..#..",
   "...##...",
   "..#..#..",
   "..#..#..",   
   "...##...",
  ] 
}

registerSprite(UpwardSparkSpawner);

class UpwardSparkSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x+1, y+1, // left,top are offset by 1,1 from spawner's left,top position
   6,6 // width,height
  )
  this.setAnimation(
   0,1808, // left,top in spritesheet
   16, 16, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setDespawnAnimation(
   256,1808, // left,top in spritesheet
   16, 4, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   0,-4, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   3, // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.  
}


class RightwardSparkSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x-7, y-1, // left,top in world, offset by -5,-1 from stamp left,top
   11,10 // width,height
  );
  this.setAnimation(
   0,1840, // left,top in spritesheet
   16, 29, // spritesheet grid size, number of grid spaces
   4 // animation timing
  );
  this.addPeriodicSpawn(
   RightwardSparkSpawn, // sprite class to spawn
   29*4, // 29 sprite frames by 3 timing frames, so once per animation loop
   10*4, // fire at 10th sprite frame
  )
  this.addPeriodicSpawn(
   RightwardSparkSpawn, // sprite class to spawn
   29*4, // 29 sprite frames by 3 timing frames, so once per animation loop
   21*4, // fire at 21st sprite frame
  )
  this.addPeriodicSpawn(
   RightwardSparkSpawn, // sprite class to spawn
   29*4, // 29 sprite frames by 3 timing frames, so once per animation loop
   0, // fire at repeat point of loop
  )
  this.suppressBlink(); // don't flicker if player crashes into this
 }
 static stamp=
  [
   ".......#...",   
   ".......##..",
   ".##.####.#.",
   "#..#......#",
   "#..#......#",   
   ".##.####.#..",
   ".......##..",
   ".......#...",
  ] 
}

registerSprite(RightwardSparkSpawner);


class RightwardSparkSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x+6, y+2, // left,top are offset by 6,2 from spawner's left,top position
   6,6 // width,height
  )
  this.setAnimation(
   0,1852, // left,top in spritesheet
   16, 16, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setDespawnAnimation(
   256,1852, // left,top in spritesheet
   16, 4, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   4,0, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   3, // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.  
}



class LeftwardSparkSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x+7, y-1, // left,top in world, offset by 7,-1 from stamp left,top
   11,10 // width,height
  );
  this.setAnimation(
   0,1872, // left,top in spritesheet
   16, 29, // spritesheet grid size, number of grid spaces
   4 // animation timing
  );
  this.addPeriodicSpawn(
   LeftwardSparkSpawn, // sprite class to spawn
   29*4, // 29 sprite frames by 3 timing frames, so once per animation loop
   17*4, // fire at 17th sprite frame
  )
  this.addPeriodicSpawn(
   LeftwardSparkSpawn, // sprite class to spawn
   29*4, // 29 sprite frames by 3 timing frames, so once per animation loop
   28*4, // fire at 28th sprite frame
  )
  this.addPeriodicSpawn(
   LeftwardSparkSpawn, // sprite class to spawn
   29*4, // 29 sprite frames by 3 timing frames, so once per animation loop
   7*4, // fire at 7th sprite frame
  )
  this.suppressBlink(); // don't flicker if player crashes into this
 }
 static stamp=
  [
   "...#.......",
   "..##.......",
   ".#.####.##.",
   "#......#..#",
   "#......#..#",
   ".#.####.##.",
   "..##.......",
   "...#......."
  ] 
}

registerSprite(LeftwardSparkSpawner);


class LeftwardSparkSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x-1, y+2, // left,top are offset by -1,2 from spawner's left,top position
   6,6 // width,height
  )
  this.setAnimation(
   0,1852, // left,top in spritesheet
   16, 16, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setDespawnAnimation(
   256,1852, // left,top in spritesheet
   16, 4, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   -4,0, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   3, // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.  
}

class BubbleSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   10,9 // width,height
  );
  this.setAnimation(
   0,1904, // left,top in spritesheet
   16, 14, // spritesheet grid size, number of grid spaces
   4 // animation timing
  );
  this.addPeriodicSpawn(
   BubbleSpawn1, // sprite class to spawn
   14*4*10, // firing pattern repeats over 9 animation loops
   0, // fire at repeat point of pattern
  )
  this.addPeriodicSpawn(BubbleSpawn2, 14*4*10, 14*4*1);
  this.addPeriodicSpawn(BubbleSpawn3, 14*4*10, 14*4*2);
  this.addPeriodicSpawn(BubbleSpawn4, 14*4*10, 14*4*3);
  this.addPeriodicSpawn(BubbleSpawn5, 14*4*10, 14*4*4);
  this.addPeriodicSpawn(BubbleSpawn6, 14*4*10, 14*4*5);
  this.addPeriodicSpawn(BubbleSpawn7, 14*4*10, 14*4*6);
  this.addPeriodicSpawn(BubbleSpawn8, 14*4*10, 14*4*7);
  this.addPeriodicSpawn(BubbleSpawn9, 14*4*10, 14*4*8);
  this.addPeriodicSpawn(BubbleSpawn10, 14*4*10, 14*4*9);  
 }
 static stamp=[
  "...####...",
  "..##..##..",
  "..#....#..",
  "..#....#..",
  "..##..##..",
  "##########",
  "#........#",
  ".##....##.",
  "...####..."
 ]
}

registerSprite(BubbleSpawner);

/* ten bubbles that differ only in trajectory; a base class can hold all the
 common values */
class AbstractBubbleSpawn extends CustomSprite {
 constructor(x,y,xv,yv,xa,ya) {
  super();
  this.setRect(
   x+2, y-1, // left,top point is two pixels right and one up from spawner's
   6,6 // width,height
  )
  this.setAnimation(
   0,1920, // left,top in spritesheet
   16, 10, // spritesheet grid size, number of grid spaces
   3 // animation timing
  )
  this.setDespawnAnimation(
   0,1936, // left,top in spritesheet
   16, 5, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   xv,yv, // x,y velocity in subpixels as defined in subclass's super call
   0,5, // accelerates downward 5 sub pixels per frame per frame
   512, // number of subpixels per pixel
  )
 }
}
class BubbleSpawn1 extends AbstractBubbleSpawn {
 constructor(x,y) {
  super(x,y, // left,top of spawner; base class is setting position from there
	750, -250 // x velocity 1500, y velocity -250
       );
 }
}
class BubbleSpawn2 extends AbstractBubbleSpawn { constructor(x,y) {
 super(x,y,-250,-375);
} }
class BubbleSpawn3 extends AbstractBubbleSpawn { constructor(x,y) {
 super(x,y,-230,-275);
} }
class BubbleSpawn4 extends AbstractBubbleSpawn { constructor(x,y) {
 super(x,y,260,-500);
} }
class BubbleSpawn5 extends AbstractBubbleSpawn { constructor(x,y) {
 super(x,y,-350,-300);
} }
class BubbleSpawn6 extends AbstractBubbleSpawn { constructor(x,y) {
 super(x,y,30,-400);
} }
class BubbleSpawn7 extends AbstractBubbleSpawn { constructor(x,y) {
 super(x,y,-730,-510);
} }
class BubbleSpawn8 extends AbstractBubbleSpawn { constructor(x,y) {
 super(x,y,600,-290);
} }
class BubbleSpawn9 extends AbstractBubbleSpawn { constructor(x,y) {
 super(x,y,-750,-260);
} }
class BubbleSpawn10 extends AbstractBubbleSpawn { constructor(x,y) {
 super(x,y,100,-520);
} }

class RedBeamSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x+3,y, // left,top in world, three pixels right from left,top of stamp
   5,5 // width,height
  );
  this.setAnimation(
   0,1952, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   1 // animation timing
  );
  this.setPatrolMove(
   0,-1, // x,y velocity in subpixels
   3, // number of subpixels per pixel
   // omit fourth argument, animation doesn't change on turnaround
  )
  this.setPeriodicPause(
   107, 20, // number of frames active, number of frames paused
   16, 1952 // left,top in spritesheet while paused
  )
  this.addPeriodicSpawn(
   RedBeamLeftSpawn, // sprite class to spawn
   107, // once per pause loop
   0 // no timing offset
  )
  this.addPeriodicSpawn(
   RedBeamRightSpawn, // sprite class to spawn
   107, // once per pause loop
   0 // no timing offset
  )  
 }
 static stamp=[
  "....###....",
  ".#.#####.#.",
  "##.#.#.#.##",
  ".#.#####.#.",
  "....###....",
 ]
}

registerSprite(RedBeamSpawner);
 
class RedBeamLeftSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x-3, y+1, // left,top offset from spawner's left,top by -3,1
   4,3 // width,height
  )
  this.setAnimation(
   0,1968, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setDespawnAnimation(
   16,1968, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   -1,0, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   1, // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.
}

class RedBeamRightSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x+3, y+1, // left,top offset from spawner's left,top by 3,1
   4,3 // width,height
  )
  this.setAnimation(
   32,1968, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setDespawnAnimation(
   48,1968, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   1,0, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   1, // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.
}


class BlueBeamSpawner extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x,y, // left,top in world
   12,12 // width,height
  );
  this.setAnimation(
   256,1952, // left,top in spritesheet
   0, 1, // spritesheet grid size, number of grid spaces
   1 // animation timing
  );
  this.setPatrolMove(
   0,1, // x,y velocity in subpixels
   4, // number of subpixels per pixel
   // omit fourth argument, animation doesn't change on turnaround
  )
  this.addPeriodicSpawn(
   BlueBeamLeftSpawn, // sprite class to spawn
   53, // game frames between consecutive shots
   0 // no timing offset
  )
  this.addPeriodicSpawn(
   BlueBeamRightSpawn, // sprite class to spawn
   53, // game frames between consecutive shots
   0 // no timing offset
  )  
 }
 static stamp=[
  "#####..#####",
  "#...####...#",
  "#..........#",
  "#..######..#",
  "##.#....#.##",
  ".#.#....#.#.",
  ".#.#....#.#.",
  "##.#....#.##",
  "#..######..#",
  "#..........#",
  "#...####...#",
  "#####..#####"
 ];
}

registerSprite(BlueBeamSpawner);

class BlueBeamLeftSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x, y+4, // left,top offset from spawner's left,top by 0,4
   4,4 // width,height
  )
  this.setAnimation(
   272,1952, // left,top in spritesheet
   8, 2, // spritesheet grid size, number of grid spaces
   3 // animation timing
  )
  this.setDespawnAnimation(
   288,1952, // left,top in spritesheet
   8, 3, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   -1,0, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   1 // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.
}

class BlueBeamRightSpawn extends CustomSprite {
 constructor(x,y) {
  super();
  this.setRect(
   x+8, y+4, // left,top offset from spawner's left,top by 8,4
   4,4 // width,height
  )
  this.setAnimation(
   272,1960, // left,top in spritesheet
   8, 2, // spritesheet grid size, number of grid spaces
   3 // animation timing
  )
  this.setDespawnAnimation(
   288,1960, // left,top in spritesheet
   8, 3, // spritesheet grid size, number of grid spaces
   1 // animation timing
  )
  this.setProjectileMove(
   1,0, // x,y velocity in subpixels
   0,0, // x,y acceleration in subpixels
   1 // number of subpixels per pixel
  )
 }
 // This is spawned by another sprite and has no stamp to register.
}

